from .py_simple_morse_code import (
    MorseCodeTranslator,
    encode_string_to_morse,
    decode_morse_to_string,
    encode_morse_to_beats,
    encode_string_to_beats,
    make_morse_visual_from_beats,
    encode_beats_to_waveform,
    play_waveform,
    play_string,
    play_morse,
    encode_string_to_waveform,
    encode_morse_to_waveform,
)
from .signal_processing import SignalProcessor
from .CONSTANTS import (
    DEFAULT_TONE,
    DEFAULT_MIN_TONE,
    DEFAULT_MAX_TONE,
    DEFAULT_SAMPLE_SIZE,
    DEFAULT_SAMPLE_RATE,
    DEFAULT_HIGH_PASS_FREQUENCY,
    DEFAULT_LOW_PASS_FREQUENCY,
    DEFAULT_NUM_DEADBEATS,
    DEFAULT_WORDS_PER_MINUTE,
    DEFAULT_STATE_DEBOUNCE,
    DURATIONS,
    UNKNOWN_CHAR,
    MORSE_CODE_LETTERS_MAP,
    MORSE_CODE_NUMBERS_MAP,
    MORSE_CODE_PUNCTUATION_MAP,
    CHAR_TO_MORSE_MAP,
    MORSE_TO_CHAR_MAP,
)
from .version import version
